#output {
	background: rgba(0, 0, 0, 0.4);
	border-radius: 0;
	border: 1px solid rgba(200, 200, 200, 0.3);
	color: #eee;
	text-shadow: none;
	box-shadow: none
}

#output:hover {
	background: rgba(92, 92, 92, 0.05)
}

#selected-output {
    background: rgba(0, 0, 0, 0.4);
	border-radius: 0;
	border: 1px solid rgba(200, 200, 255, 0.6);
	color: #eee;
	text-shadow: none;
	box-shadow: none;
    font-weight: bold
}

#inactive-output {
    background: rgba(0, 0, 0, 0.4);
	border-radius: 0;
	border: none rgba(200, 200, 255, 0.6);
	color: #333;
	text-shadow: none;
	box-shadow: none;
    font-weight: bold
}

#wrapper {
    background: rgba(0, 0, 0, 0.5);
}

#indicator {
    background: rgba(0, 0, 0, 0.8);
	border-radius: 0;
	border: 1px solid rgba(200, 200, 255, 0.6);
	color: #eee
}

#indicator-label {
    font-size: 24px
}